# -*- coding: utf-8 -*-
# Copyright 2014 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from twisted.internet import defer

from synapse.api.errors import SynapseError, AuthError, CodeMessageException
from synapse.api.constants import Membership
from synapse.api.events.room import RoomMemberEvent

from ._base import BaseHandler

import logging


logger = logging.getLogger(__name__)


class ProfileHandler(BaseHandler):

    def __init__(self, hs):
        super(ProfileHandler, self).__init__(hs)

        self.federation = hs.get_replication_layer()
        self.federation.register_query_handler(
            "profile", self.on_profile_query
        )

        distributor = hs.get_distributor()
        self.distributor = distributor

        distributor.observe("registered_user", self.registered_user)

        distributor.observe(
            "collect_presencelike_data", self.collect_presencelike_data
        )

    def registered_user(self, user):
        self.store.create_profile(user.localpart)

    @defer.inlineCallbacks
    def get_displayname(self, target_user):
        if target_user.is_mine:
            displayname = yield self.store.get_profile_displayname(
                target_user.localpart
            )

            defer.returnValue(displayname)
        else:
            try:
                result = yield self.federation.make_query(
                    destination=target_user.domain,
                    query_type="profile",
                    args={
                        "user_id": target_user.to_string(),
                        "field": "displayname",
                    }
                )
            except CodeMessageException as e:
                if e.code != 404:
                    logger.exception("Failed to get displayname")

                raise
            except:
                logger.exception("Failed to get displayname")
            else:
                defer.returnValue(result["displayname"])

    @defer.inlineCallbacks
    def set_displayname(self, target_user, auth_user, new_displayname):
        """target_user is the user whose displayname is to be changed;
        auth_user is the user attempting to make this change."""
        if not target_user.is_mine:
            raise SynapseError(400, "User is not hosted on this Home Server")

        if target_user != auth_user:
            raise AuthError(400, "Cannot set another user's displayname")

        yield self.store.set_profile_displayname(
            target_user.localpart, new_displayname
        )

        yield self.distributor.fire(
            "changed_presencelike_data", target_user, {
                "displayname": new_displayname,
            }
        )

        yield self._update_join_states(target_user)

    @defer.inlineCallbacks
    def get_avatar_url(self, target_user):
        if target_user.is_mine:
            avatar_url = yield self.store.get_profile_avatar_url(
                target_user.localpart
            )

            defer.returnValue(avatar_url)
        else:
            try:
                result = yield self.federation.make_query(
                    destination=target_user.domain,
                    query_type="profile",
                    args={
                        "user_id": target_user.to_string(),
                        "field": "avatar_url",
                    }
                )
            except CodeMessageException as e:
                if e.code != 404:
                    logger.exception("Failed to get avatar_url")
                raise
            except:
                logger.exception("Failed to get avatar_url")

            defer.returnValue(result["avatar_url"])

    @defer.inlineCallbacks
    def set_avatar_url(self, target_user, auth_user, new_avatar_url):
        """target_user is the user whose avatar_url is to be changed;
        auth_user is the user attempting to make this change."""
        if not target_user.is_mine:
            raise SynapseError(400, "User is not hosted on this Home Server")

        if target_user != auth_user:
            raise AuthError(400, "Cannot set another user's avatar_url")

        yield self.store.set_profile_avatar_url(
            target_user.localpart, new_avatar_url
        )

        yield self.distributor.fire(
            "changed_presencelike_data", target_user, {
                "avatar_url": new_avatar_url,
            }
        )

        yield self._update_join_states(target_user)

    @defer.inlineCallbacks
    def collect_presencelike_data(self, user, state):
        if not user.is_mine:
            defer.returnValue(None)

        (displayname, avatar_url) = yield defer.gatherResults([
            self.store.get_profile_displayname(user.localpart),
            self.store.get_profile_avatar_url(user.localpart),
        ])

        state["displayname"] = displayname
        state["avatar_url"] = avatar_url

        defer.returnValue(None)

    @defer.inlineCallbacks
    def on_profile_query(self, args):
        user = self.hs.parse_userid(args["user_id"])
        if not user.is_mine:
            raise SynapseError(400, "User is not hosted on this Home Server")

        just_field = args.get("field", None)

        response = {}

        if just_field is None or just_field == "displayname":
            response["displayname"] = yield self.store.get_profile_displayname(
                user.localpart
            )

        if just_field is None or just_field == "avatar_url":
            response["avatar_url"] = yield self.store.get_profile_avatar_url(
                user.localpart
            )

        defer.returnValue(response)

    @defer.inlineCallbacks
    def _update_join_states(self, user):
        if not user.is_mine:
            return

        joins = yield self.store.get_rooms_for_user_where_membership_is(
            user.to_string(),
            [Membership.JOIN],
        )

        for j in joins:
            snapshot = yield self.store.snapshot_room(
                j.room_id, j.state_key, RoomMemberEvent.TYPE,
                j.state_key
            )

            content = {
                "membership": j.content["membership"],
                "prev": j.content["membership"],
            }

            yield self.distributor.fire(
                "collect_presencelike_data", user, content
            )

            new_event = self.event_factory.create_event(
                etype=j.type,
                room_id=j.room_id,
                state_key=j.state_key,
                content=content,
                user_id=j.state_key,
            )

            yield self.state_handler.handle_new_event(new_event, snapshot)
            yield self._on_new_room_event(new_event, snapshot)
